import importlib.metadata

from .base import (
    LightningIRClassFactory,
    LightningIRConfig,
    LightningIRConfigClassFactory,
    LightningIRModel,
    LightningIRModelClassFactory,
    LightningIRModule,
    LightningIROutput,
    LightningIRTokenizer,
    LightningIRTokenizerClassFactory,
)
from .bi_encoder import (
    BiEncoderConfig,
    BiEncoderEmbedding,
    BiEncoderModel,
    BiEncoderModule,
    BiEncoderOutput,
    BiEncoderTokenizer,
    MultiVectorBiEncoderConfig,
    MultiVectorBiEncoderModel,
    SingleVectorBiEncoderConfig,
    SingleVectorBiEncoderModel,
)
from .callbacks import IndexCallback, RankCallback, RegisterLocalDatasetCallback, ReRankCallback, SearchCallback
from .cross_encoder import (
    CrossEncoderConfig,
    CrossEncoderModel,
    CrossEncoderModule,
    CrossEncoderOutput,
    CrossEncoderTokenizer,
)
from .data import (
    DocDataset,
    DocSample,
    IndexBatch,
    LightningIRDataModule,
    QueryDataset,
    QuerySample,
    RankBatch,
    RankSample,
    RunDataset,
    SearchBatch,
    TrainBatch,
    TupleDataset,
)
from .data.external_datasets import _register_external_datasets
from .loss import (
    ApproxMRR,
    ApproxNDCG,
    ApproxRankMSE,
    ConstantMarginMSE,
    ContrastiveLocalLoss,
    FLOPSRegularization,
    InBatchCrossEntropy,
    InfoNCE,
    KLDivergence,
    L1Regularization,
    L2Regularization,
    PearsonCorrelation,
    RankNet,
    ScoreBasedInBatchCrossEntropy,
    ScoreBasedInBatchLossFunction,
    SupervisedMarginMSE,
)
from .main import LightningIRTrainer, LightningIRWandbLogger
from .models.register_external_models import _register_external_models
from .models.register_internal_models import _register_internal_models
from .retrieve import (
    FaissFlatIndexConfig,
    FaissFlatIndexer,
    FaissIVFIndexConfig,
    FaissIVFIndexer,
    FaissIVFPQIndexConfig,
    FaissIVFPQIndexer,
    FaissPQIndexConfig,
    FaissPQIndexer,
    FaissSearchConfig,
    FaissSearcher,
    IndexConfig,
    Indexer,
    SearchConfig,
    Searcher,
    SeismicIndexConfig,
    SeismicIndexer,
    SeismicSearchConfig,
    SeismicSearcher,
    TorchDenseIndexConfig,
    TorchDenseIndexer,
    TorchDenseSearchConfig,
    TorchDenseSearcher,
    TorchSparseIndexConfig,
    TorchSparseIndexer,
    TorchSparseSearchConfig,
    TorchSparseSearcher,
)
from .schedulers import (
    ConstantLRSchedulerWithLinearWarmup,
    GenericConstantSchedulerWithLinearWarmup,
    GenericConstantSchedulerWithQuadraticWarmup,
    GenericLinearSchedulerWithLinearWarmup,
    LinearLRSchedulerWithLinearWarmup,
    WarmupLRScheduler,
)

__version__ = importlib.metadata.version("lightning-ir")

_register_internal_models()
_register_external_models()
_register_external_datasets()


__all__ = [
    "ApproxMRR",
    "ApproxNDCG",
    "ApproxRankMSE",
    "BiEncoderConfig",
    "BiEncoderEmbedding",
    "BiEncoderModel",
    "BiEncoderModule",
    "BiEncoderOutput",
    "BiEncoderTokenizer",
    "ConstantLRSchedulerWithLinearWarmup",
    "ConstantMarginMSE",
    "ContrastiveLocalLoss",
    "CrossEncoderConfig",
    "CrossEncoderModel",
    "CrossEncoderModule",
    "CrossEncoderOutput",
    "CrossEncoderTokenizer",
    "DocDataset",
    "DocSample",
    "FaissFlatIndexConfig",
    "FaissFlatIndexer",
    "FaissIVFIndexConfig",
    "FaissIVFIndexer",
    "FaissIVFPQIndexConfig",
    "FaissIVFPQIndexer",
    "FaissPQIndexConfig",
    "FaissPQIndexer",
    "FaissSearchConfig",
    "FaissSearcher",
    "FLOPSRegularization",
    "GenericConstantSchedulerWithLinearWarmup",
    "GenericConstantSchedulerWithQuadraticWarmup",
    "GenericLinearSchedulerWithLinearWarmup",
    "InBatchCrossEntropy",
    "IndexBatch",
    "IndexCallback",
    "IndexConfig",
    "Indexer",
    "InfoNCE",
    "KLDivergence",
    "L1Regularization",
    "L2Regularization",
    "LightningIRClassFactory",
    "LightningIRConfig",
    "LightningIRConfigClassFactory",
    "LightningIRDataModule",
    "LightningIRModel",
    "LightningIRModelClassFactory",
    "LightningIRModule",
    "LightningIROutput",
    "LightningIRTokenizer",
    "LightningIRTokenizerClassFactory",
    "LightningIRTrainer",
    "LightningIRWandbLogger",
    "LinearLRSchedulerWithLinearWarmup",
    "MultiVectorBiEncoderConfig",
    "MultiVectorBiEncoderModel",
    "PearsonCorrelation",
    "QueryDataset",
    "QuerySample",
    "RankBatch",
    "RankCallback",
    "RankNet",
    "RankSample",
    "RegisterLocalDatasetCallback",
    "ReRankCallback",
    "RunDataset",
    "ScoreBasedInBatchCrossEntropy",
    "ScoreBasedInBatchLossFunction",
    "SearchBatch",
    "SearchCallback",
    "SearchConfig",
    "Searcher",
    "SeismicIndexConfig",
    "SeismicIndexer",
    "SeismicSearchConfig",
    "SeismicSearcher",
    "SingleVectorBiEncoderConfig",
    "SingleVectorBiEncoderModel",
    "SupervisedMarginMSE",
    "TorchDenseIndexConfig",
    "TorchDenseIndexer",
    "TorchDenseSearchConfig",
    "TorchDenseSearcher",
    "TorchSparseIndexConfig",
    "TorchSparseIndexer",
    "TorchSparseSearchConfig",
    "TorchSparseSearcher",
    "TrainBatch",
    "TupleDataset",
    "WarmupLRScheduler",
]



from typing import Literal

import torch


class Pooler(torch.nn.Module):
    """Applies pooling to the embeddings based on the pooling strategy defined in the configuration."""

    def __init__(self, pooling_strategy: Literal["first", "mean", "max", "sum"]) -> None:
        """Initializes the pooler.

        Args:
            pooling_strategy (Literal['first', 'mean', 'max', 'sum']): Pooling strategy to aggregate the
                contextualized embeddings into a single vector.
        """
        super().__init__()
        self.pooling_strategy = pooling_strategy

    def forward(self, embeddings: torch.Tensor, attention_mask: torch.Tensor | None = None) -> torch.Tensor:
        """Applies optional pooling to the embeddings.

        Args:
            embeddings (torch.Tensor): Query or document embeddings
            attention_mask (torch.Tensor | None): Query or document attention mask
        Returns:
            torch.Tensor: (Optionally) pooled embeddings.
        Raises:
            ValueError: If an unknown pooling strategy is passed.
        """
        if self.pooling_strategy is None:
            return embeddings
        if self.pooling_strategy == "first":
            return embeddings[:, [0]]
        if self.pooling_strategy in ("sum", "mean"):
            if attention_mask is not None:
                embeddings = embeddings * attention_mask.unsqueeze(-1)
            embeddings = embeddings.sum(dim=1, keepdim=True)
            if self.pooling_strategy == "mean":
                if attention_mask is not None:
                    embeddings = embeddings / attention_mask.sum(dim=1, keepdim=True).unsqueeze(-1)
            return embeddings
        if self.pooling_strategy == "max":
            if attention_mask is not None:
                embeddings = embeddings.masked_fill(~attention_mask.bool().unsqueeze(-1), float("-inf"))
            return embeddings.amax(dim=1, keepdim=True)
        raise ValueError(f"Unknown pooling strategy: {self.pooling_strategy}")


class Sparsifier(torch.nn.Module):
    """Applies sparsification to the embeddings based on the sparsification strategy defined in the configuration."""

    def __init__(self, sparsification_strategy: Literal["relu", "relu_log", "relu_2xlog"] | None) -> None:
        """Initializes the sparsifier.

        Args:
            sparsification_strategy (Literal['relu', 'relu_log', 'relu_2xlog'] | None): Which sparsification strategy
                to apply.
        """
        super().__init__()
        self.sparsification_strategy = sparsification_strategy

    def forward(self, embeddings: torch.Tensor) -> torch.Tensor:
        """Applies optional sparsification to the embeddings.

        Args:
            embeddings (torch.Tensor): Query or document embeddings
        Returns:
            torch.Tensor: (Optionally) sparsified embeddings.
        Raises:
            ValueError: If an unknown sparsification strategy is passed.
        """
        if self.sparsification_strategy is None:
            return embeddings
        if self.sparsification_strategy == "relu":
            return torch.relu(embeddings)
        if self.sparsification_strategy == "relu_log":
            return torch.log1p(torch.relu(embeddings))
        if self.sparsification_strategy == "relu_2xlog":
            return torch.log1p(torch.log1p(torch.relu(embeddings)))
        raise ValueError(f"Unknown sparsification strategy: {self.sparsification_strategy}")



from functools import wraps
from typing import Callable

import torch


def _batch_elementwise_scoring(
    similarity_function: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
    """Helper function to batch similarity functions to avoid memory issues with large batch sizes or high numbers
    of documents per query."""
    BATCH_SIZE = 16384

    @wraps(similarity_function)
    def batch_similarity_function(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """x and y have shape (N, ..., D) where ... are broadcastable dimensions and D is the embedding dimension."""
        if x.shape[0] <= BATCH_SIZE:
            return similarity_function(x, y)
        out = torch.empty(x.shape[0], x.shape[1], y.shape[2], device=x.device, dtype=x.dtype)
        for i in range(0, x.shape[0], BATCH_SIZE):
            out[i : i + BATCH_SIZE] = similarity_function(x[i : i + BATCH_SIZE], y[i : i + BATCH_SIZE])
        return out

    return batch_similarity_function


def _batch_pairwise_scoring(
    similarity_function: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
    """Helper function to batch similarity functions to avoid memory issues with large batch sizes or high numbers
    of documents per query."""
    BATCH_SIZE = 8192

    @wraps(similarity_function)
    def batch_similarity_function(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """x has shape (N, D) and y has shape (M, D)"""
        if x.shape[0] <= BATCH_SIZE and y.shape[0] <= BATCH_SIZE:
            return similarity_function(x, y)
        out = torch.empty(x.shape[0], y.shape[0], device=x.device, dtype=x.dtype)
        for i in range(0, x.shape[0], BATCH_SIZE):
            for j in range(0, y.shape[0], BATCH_SIZE):
                out[i : i + BATCH_SIZE, j : j + BATCH_SIZE] = similarity_function(
                    x[i : i + BATCH_SIZE], y[j : j + BATCH_SIZE]
                )
        return out

    return batch_similarity_function



"""
Lightning IR module for native cross-encoder models.

This module provides the classes, configurations and tokenizer for various cross-encoders in the Lightning IR framework.
"""

from .mono import MonoConfig, MonoModel
from .set_encoder import SetEncoderConfig, SetEncoderModel, SetEncoderTokenizer

__all__ = [
    "MonoConfig",
    "MonoModel",
    "SetEncoderConfig",
    "SetEncoderModel",
    "SetEncoderTokenizer",
]



"""
Module module for bi-encoder models.

This module defines the Lightning IR module class used to implement bi-encoder models.
"""

from __future__ import annotations

from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any

import torch
from transformers import BatchEncoding, PreTrainedModel

from ..base import LightningIRModule, LightningIROutput
from ..data import IndexBatch, RankBatch, SearchBatch, TrainBatch
from ..loss.base import EmbeddingLossFunction, LossFunction, ScoringLossFunction
from ..loss.in_batch import InBatchLossFunction
from .bi_encoder_config import BiEncoderConfig
from .bi_encoder_model import BiEncoderEmbedding, BiEncoderModel, BiEncoderOutput
from .bi_encoder_tokenizer import BiEncoderTokenizer

if TYPE_CHECKING:
    from ..retrieve import SearchConfig, Searcher


class BiEncoderModule(LightningIRModule):
    def __init__(
        self,
        model_name_or_path: str | None = None,
        config: BiEncoderConfig | None = None,
        model: BiEncoderModel | None = None,
        BackboneModel: type[PreTrainedModel] | None = None,
        loss_functions: Sequence[LossFunction | tuple[LossFunction, float]] | None = None,
        evaluation_metrics: Sequence[str] | None = None,
        index_dir: Path | None = None,
        search_config: SearchConfig | None = None,
        model_kwargs: Mapping[str, Any] | None = None,
    ):
        """:class:`.LightningIRModule` for bi-encoder models. It contains a :class:`.BiEncoderModel` and a
        :class:`.BiEncoderTokenizer` and implements the training, validation, and testing steps for the model.

        .. _ir-measures: https://ir-measur.es/en/latest/index.html

        Args:
            model_name_or_path (str | None): Name or path of backbone model or fine-tuned Lightning IR model.
                Defaults to None.
            config (BiEncoderConfig | None): BiEncoderConfig to apply when loading from backbone model.
                Defaults to None.
            model (BiEncoderModel | None): Already instantiated BiEncoderModel. Defaults to None.
            BackboneModel (type[PreTrainedModel] | None): Huggingface PreTrainedModel class to use as backbone
                instead of the default AutoModel. Defaults to None.
            loss_functions (Sequence[LossFunction | tuple[LossFunction, float]] | None):
                Loss functions to apply during fine-tuning, optional loss weights can be provided per loss function
                Defaults to None.
            evaluation_metrics (Sequence[str] | None): Metrics corresponding to ir-measures_ measure strings
                to apply during validation or testing. Defaults to None.
            index_dir (Path | None): Path to an index used for retrieval. Defaults to None.
            search_config (SearchConfig | None): Configuration to use during retrieval. Defaults to None.
            model_kwargs (Mapping[str, Any] | None): Additional keyword arguments to pass to `from_pretrained`
                when loading a model. Defaults to None.
        """
        super().__init__(
            model_name_or_path=model_name_or_path,
            config=config,
            model=model,
            BackboneModel=BackboneModel,
            loss_functions=loss_functions,
            evaluation_metrics=evaluation_metrics,
            model_kwargs=model_kwargs,
        )
        self.model: BiEncoderModel
        self.config: BiEncoderConfig
        self.tokenizer: BiEncoderTokenizer
        if len(self.tokenizer) > self.config.vocab_size:
            self.model.resize_token_embeddings(len(self.tokenizer), 8)
        self._searcher = None
        self.search_config = search_config
        self.index_dir = index_dir

    @property
    def searcher(self) -> Searcher | None:
        """Searcher used for retrieval if `index_dir` and `search_config` are set.

        Returns:
            Searcher: Searcher class.
        """
        return self._searcher

    @searcher.setter
    def searcher(self, searcher: Searcher):
        self._searcher = searcher

    def _init_searcher(self) -> None:
        if self.search_config is not None and self.index_dir is not None:
            self.searcher = self.search_config.search_class(self.index_dir, self.search_config, self)

    def on_test_start(self) -> None:
        """Called at the beginning of testing. Initializes the searcher if `index_dir` and `search_config` are set."""
        self._init_searcher()
        return super().on_test_start()

    def forward(self, batch: RankBatch | IndexBatch | SearchBatch) -> BiEncoderOutput:
        """Runs a forward pass of the model on a batch of data. The output will vary depending on the type of batch. If
        the batch is a :class`.RankBatch`, query and document embeddings are computed and the relevance score is the
        similarity between the two embeddings. If the batch is an :class:`.IndexBatch`, only document embeddings
        are comuputed. If the batch is a :class:`.SearchBatch`, only query embeddings are computed and
        the model will additionally retrieve documents if :attr:`.searcher` is set.

        Args:
            batch (RankBatch | IndexBatch | SearchBatch): Input batch containing queries and/or documents.
        Returns:
            BiEncoderOutput: Output of the model.
        Raises:
            ValueError: If the input batch contains neither queries nor documents.
        """
        queries = getattr(batch, "queries", None)
        docs = getattr(batch, "docs", None)
        num_docs = None
        if isinstance(batch, RankBatch):
            num_docs = None if docs is None else [len(d) for d in docs]
            docs = [d for nested in docs for d in nested] if docs is not None else None
        encodings = self.prepare_input(queries, docs, num_docs)

        if not encodings:
            raise ValueError("No encodings were generated.")
        output = self.model.forward(
            encodings.get("query_encoding", None), encodings.get("doc_encoding", None), num_docs
        )
        doc_ids = getattr(batch, "doc_ids", None)
        if doc_ids is not None and output.doc_embeddings is not None:
            output.doc_embeddings.ids = doc_ids
        query_ids = getattr(batch, "query_ids", None)
        if query_ids is not None and output.query_embeddings is not None:
            output.query_embeddings.ids = query_ids
        if isinstance(batch, SearchBatch) and self.searcher is not None:
            scores, doc_ids = self.searcher.search(output)
            output.scores = scores
            if output.doc_embeddings is not None:
                output.doc_embeddings.ids = [doc_id for _doc_ids in doc_ids for doc_id in _doc_ids]
            batch.doc_ids = doc_ids
        return output

    def score(self, queries: Sequence[str] | str, docs: Sequence[Sequence[str]] | Sequence[str]) -> BiEncoderOutput:
        """Computes relevance scores for queries and documents.

        Args:
            queries (Sequence[str] | str): Queries to score.
            docs (Sequence[Sequence[str]] | Sequence[str]): Documents to score.
        Returns:
            BiEncoderOutput: Output of the model.
        """
        return super().score(queries, docs)

    def _compute_losses(self, batch: TrainBatch, output: BiEncoderOutput) -> list[torch.Tensor]:
        """Computes the losses for a training batch."""
        if self.loss_functions is None:
            raise ValueError("Loss function is not set")

        if (
            batch.targets is None
            or output.query_embeddings is None
            or output.doc_embeddings is None
            or output.scores is None
        ):
            raise ValueError(
                "targets, scores, query_embeddings, and doc_embeddings must be set in the output and batch"
            )

        num_queries = len(batch.queries)
        output.scores = output.scores.view(num_queries, -1)
        batch.targets = batch.targets.view(*output.scores.shape, -1)
        losses = []
        for loss_function, _ in self.loss_functions:
            if isinstance(loss_function, InBatchLossFunction):
                pos_idcs, neg_idcs = loss_function.get_ib_idcs(output, batch)
                ib_doc_embeddings = self._get_ib_doc_embeddings(output.doc_embeddings, pos_idcs, neg_idcs, num_queries)
                ib_scores = self.model.score(
                    BiEncoderOutput(query_embeddings=output.query_embeddings, doc_embeddings=ib_doc_embeddings)
                ).scores
                if ib_scores is None:
                    raise ValueError("In-batch scores cannot be None")
                ib_scores = ib_scores.view(num_queries, -1)
                losses.append(loss_function.compute_loss(LightningIROutput(ib_scores)))
            elif isinstance(loss_function, EmbeddingLossFunction):
                losses.append(loss_function.compute_loss(output))
            elif isinstance(loss_function, ScoringLossFunction):
                losses.append(loss_function.compute_loss(output, batch))
            else:
                raise ValueError(f"Unknown loss function type {loss_function.__class__.__name__}")
        if self.config.sparsification_strategy is not None:
            query_num_nonzero = (
                torch.nonzero(output.query_embeddings.embeddings).shape[0] / output.query_embeddings.embeddings.shape[0]
            )
            doc_num_nonzero = (
                torch.nonzero(output.doc_embeddings.embeddings).shape[0] / output.doc_embeddings.embeddings.shape[0]
            )
            self.log("query_num_nonzero", query_num_nonzero)
            self.log("doc_num_nonzero", doc_num_nonzero)
        return losses

    def _get_ib_doc_embeddings(
        self,
        embeddings: BiEncoderEmbedding,
        pos_idcs: torch.Tensor,
        neg_idcs: torch.Tensor,
        num_queries: int,
    ) -> BiEncoderEmbedding:
        """Gets the in-batch document embeddings for a training batch."""
        _, num_embs, emb_dim = embeddings.embeddings.shape
        ib_embeddings = torch.cat(
            [
                embeddings.embeddings[pos_idcs].view(num_queries, -1, num_embs, emb_dim),
                embeddings.embeddings[neg_idcs].view(num_queries, -1, num_embs, emb_dim),
            ],
            dim=1,
        ).view(-1, num_embs, emb_dim)
        if embeddings.scoring_mask is None:
            ib_scoring_mask = None
        else:
            ib_scoring_mask = torch.cat(
                [
                    embeddings.scoring_mask[pos_idcs].view(num_queries, -1, num_embs),
                    embeddings.scoring_mask[neg_idcs].view(num_queries, -1, num_embs),
                ],
                dim=1,
            ).view(-1, num_embs)
        if embeddings.encoding is None:
            ib_encoding = None
        else:
            ib_encoding = {}
            for key, value in embeddings.encoding.items():
                seq_len = value.shape[-1]
                ib_encoding[key] = torch.cat(
                    [value[pos_idcs].view(num_queries, -1, seq_len), value[neg_idcs].view(num_queries, -1, seq_len)],
                    dim=1,
                ).view(-1, seq_len)
            ib_encoding = BatchEncoding(ib_encoding)
        return BiEncoderEmbedding(ib_embeddings, ib_scoring_mask, ib_encoding)

    def validation_step(
        self,
        batch: TrainBatch | IndexBatch | SearchBatch | RankBatch,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> BiEncoderOutput:
        """Handles the validation step for the model.

        Args:
            batch (TrainBatch | IndexBatch | SearchBatch | RankBatch): Batch of validation or testing data.
            batch_idx (int): Index of the batch.
            dataloader_idx (int | None): Index of the dataloader. Defaults to 0.
        Returns:
            BiEncoderOutput: Output of the model.
        """
        if isinstance(batch, IndexBatch):
            return self.forward(batch)
        if isinstance(batch, (RankBatch, TrainBatch, SearchBatch)):
            return super().validation_step(batch, batch_idx, dataloader_idx)
        raise ValueError(f"Unknown batch type {type(batch)}")



"""
Neural sorting-based loss functions for the Lightning IR framework.

This module contains loss functions that use neural sorting techniques
to compute differentiable ranking-based losses.
"""

from __future__ import annotations

import torch

from .base import ListwiseLossFunction


class NeuralLossFunction(ListwiseLossFunction):
    """Base class for neural loss functions that compute ranks from scores using neural sorting."""

    # TODO add neural loss functions

    def __init__(self, temperature: float = 1, tol: float = 1e-5, max_iter: int = 50) -> None:
        """Initialize the NeuralLossFunction.

        Args:
            temperature (float): Temperature parameter for scaling the scores. Defaults to 1.
            tol (float): Tolerance for convergence. Defaults to 1e-5.
            max_iter (int): Maximum number of iterations for convergence. Defaults to 50.
        """
        super().__init__()
        self.temperature = temperature
        self.tol = tol
        self.max_iter = max_iter

    def neural_sort(self, scores: torch.Tensor) -> torch.Tensor:
        """Compute the neural sort permutation matrix from scores.

        Args:
            scores (torch.Tensor): The input scores tensor.
        Returns:
            torch.Tensor: The computed permutation matrix.
        """
        # https://github.com/ermongroup/neuralsort/blob/master/pytorch/neuralsort.py
        scores = scores.unsqueeze(-1)
        dim = scores.shape[1]
        one = torch.ones((dim, 1), device=scores.device)

        A_scores = torch.abs(scores - scores.permute(0, 2, 1))
        B = torch.matmul(A_scores, torch.matmul(one, torch.transpose(one, 0, 1)))
        scaling = dim + 1 - 2 * (torch.arange(dim, device=scores.device) + 1)
        C = torch.matmul(scores, scaling.to(scores).unsqueeze(0))

        P_max = (C - B).permute(0, 2, 1)
        P_hat = torch.nn.functional.softmax(P_max / self.temperature, dim=-1)

        P_hat = self.sinkhorn_scaling(P_hat)

        return P_hat

    def sinkhorn_scaling(self, mat: torch.Tensor) -> torch.Tensor:
        """Apply Sinkhorn scaling to the permutation matrix.

        Args:
            mat (torch.Tensor): The input permutation matrix.
        Returns:
            torch.Tensor: The scaled permutation matrix.
        """
        # https://github.com/allegro/allRank/blob/master/allrank/models/losses/loss_utils.py#L8
        idx = 0
        while True:
            if (
                torch.max(torch.abs(mat.sum(dim=2) - 1.0)) < self.tol
                and torch.max(torch.abs(mat.sum(dim=1) - 1.0)) < self.tol
            ) or idx > self.max_iter:
                break
            mat = mat / mat.sum(dim=1, keepdim=True).clamp(min=1e-12)
            mat = mat / mat.sum(dim=2, keepdim=True).clamp(min=1e-12)
            idx += 1

        return mat

    def get_sorted_targets(self, scores: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """Get the sorted targets based on the neural sort permutation matrix.

        Args:
            scores (torch.Tensor): The input scores tensor.
            targets (torch.Tensor): The targets tensor.
        Returns:
            torch.Tensor: The sorted targets tensor.
        """
        permutation_matrix = self.neural_sort(scores)
        pred_sorted_targets = torch.matmul(permutation_matrix, targets[..., None].to(permutation_matrix)).squeeze(-1)
        return pred_sorted_targets



"""
Lightning IR module for loss functions.

This module provides various loss functions used in the Lightning IR framework, including in-batch losses, pairwise
losses, listwise losses, approximate losses, and regularization techniques.
"""

from .approximate import ApproxMRR, ApproxNDCG, ApproxRankMSE
from .base import (
    EmbeddingLossFunction,
    InBatchLossFunction,
    LossFunction,
    RegularizationLossFunction,
    ScoringLossFunction,
)
from .embedding import ContrastiveLocalLoss
from .in_batch import InBatchCrossEntropy, ScoreBasedInBatchCrossEntropy, ScoreBasedInBatchLossFunction
from .listwise import InfoNCE, KLDivergence, PearsonCorrelation
from .pairwise import ConstantMarginMSE, RankNet, SupervisedMarginMSE
from .regularization import FLOPSRegularization, L1Regularization, L2Regularization

__all__ = [
    "ApproxMRR",
    "ApproxNDCG",
    "ApproxRankMSE",
    "ConstantMarginMSE",
    "ContrastiveLocalLoss",
    "EmbeddingLossFunction",
    "FLOPSRegularization",
    "InBatchCrossEntropy",
    "InBatchLossFunction",
    "InfoNCE",
    "KLDivergence",
    "L1Regularization",
    "L2Regularization",
    "LossFunction",
    "PearsonCorrelation",
    "RankNet",
    "RegularizationLossFunction",
    "ScoreBasedInBatchCrossEntropy",
    "ScoreBasedInBatchLossFunction",
    "ScoringLossFunction",
    "SupervisedMarginMSE",
]



from .ir_datasets_utils import ParquetScoredDocs, register_new_dataset


def register_rank_distillm():
    base_url = "https://huggingface.co/datasets/webis/rank-distillm/resolve/main"

    md5_hashes = {
        "rank-distillm-rankzephyr": (
            "__rankzephyr-colbert-10000-sampled-100__msmarco-passage-train-judged.parquet",
            "38f69a3c8a5ed21c639e882a6c2eff7c",
        ),
        "rank-distillm-set-encoder": (
            "__set-encoder-colbert-all-100__msmarco-passage-train-judged.parquet",
            "a47206da7dc551e3ebd4e5b6866be78a",
        ),
        "rank-distillm-monoelectra": (
            "__monoelectra-colbert-all-100__msmarco-passage-train-judged.parquet",
            "6561f33476a6c8408737f38ea85c848f",
        ),
        "rank-distillm-colbert": (
            "__colbert__msmarco-passage-train-judged.parquet",
            "1e927d52af085516bf5a3de2865809d5",
        ),
    }
    for name, (file_name, expected_md5) in md5_hashes.items():
        dlc_contents = {
            "url": f"{base_url}/{file_name}",
            "expected_md5": expected_md5,
            "cache_path": f"msmarco-passage/train/{name}.parquet",
        }
        register_new_dataset(
            f"msmarco-passage/train/{name}",
            docs="msmarco-passage",
            queries="msmarco-passage/train",
            qrels="msmarco-passage/train",
            scoreddocs=dlc_contents,
            ScoreddocsType=ParquetScoredDocs,
        )
